#\file callbacksFatigueDNV.py
#
#----------------------------------------------------------------------------------
#
#   Callbacks for details view property behaviour
#   
#----------------------------------------------------------------------------------
#----------------------------------------------------------------------------------
#
# author: Tone Wermundsen
# Last update: 18.12.2014
# ansys v15 ACT 
#
#-----------------------------------------------------------------------------------

#----------------------------------
#
#	Onclick callbacks
#
#----------------------------------

"""Opens the documentation pdf file"""
def openAboutDNVFatigue(analysis):
	msg = "DNV Fatigue Toolkit \n"
	msg += "Developed by Tone Wermundsen as part of Master Thesis \n"
	msg += "in collaboration with NTNU and EDRMedeso \n"
	msg += "Fall 2014"
	MessageBox.Show(ExtAPI.UserInterface.MainWindow, msg, 'About DNV Fatigue Toolkit', MessageBoxType.Info, MessageBoxButtons.OK)

"""Opens window of information about the toolbar"""
def openHelpDNVFatigue(analysis):
	sourceDirectory = ExtAPI.ExtensionManager.CurrentExtension.InstallDir
	helpFile = System.IO.Path.Combine(System.IO.Path.Combine(sourceDirectory, 'documents'), 'master_ToneWermundsen.pdf')
	try:
		System.Diagnostics.Process.Start(helpFile)
	except:
		msg = "Could not find help file " + str(helpFile)
		MessageBox.Show(ExtAPI.UserInterface.MainWindow, msg, 'Error', MessageBoxType.Error, MessageBoxButtons.OK)	 

#----------------------------------
#
# Canadd callbacks
#		
#----------------------------------

"""Checks if result object can be added to analysis be checking the analysistype"""
def canAddFatigueDNV(currentAnalysis, myUserObject):
	dimension = ExtAPI.DataModel.GeoData.Assemblies[0].Dimension
	if(currentAnalysis.AnalysisType == 'Transient' or currentAnalysis.AnalysisType == 'Static') and dimension == 3 and currentAnalysis.PhysicsType == 'Mechanical':
		return True
	else:
		return False			

#----------------------------------
#
# Isvalid callbacks
#
#----------------------------------
 
"""hinders user to select snCurve HighStrengthSteel together with environment Seawater Free Corrosion"""
def isNotSFCandH(result, prop):
	if prop.Name == "environment":
		curve = prop.ParentPropertyGroup.Properties["snCurve"].ValueString
	elif prop.Name == "environmentColumn":
		curve = prop.ParentPropertyGroup.Properties["curveColumn"].ValueString
	if prop.Value == 'Seawater Free Corrosion' and curve == 'HighStrengthSteel':
		return False
	else:
		return True	

"""Validates that the referenceThickness is not less than 1.0"""
def validateReferenceThickness(result, prop):
	if result.Properties["snClass"].Properties["snEqual"].ValueString == 'Yes':
		plateThickness = prop.ParentPropertyGroup.Properties["plateThickness"].Value
	elif result.Properties["snClass"].Properties["snEqual"].ValueString == 'No': 
		plateThickness = prop.ParentPropertyGroup.Properties["plateThicknessColumn"].Value	
	try:
		if plateThickness/prop.Value < 1:
			return False
		else:
			return True	
	except:
		return True		
"""Validates that temperature given is larger than 100 Celsius"""
def validateHighTemperature(result, prop):
	minTemp = units.ConvertToUserUnit(ExtAPI, 100, 'C', 'Temperature')
	maxTemp = units.ConvertToUserUnit(ExtAPI, 786.5, 'C', 'Temperature')
	if prop.Value < minTemp:
		return False
	if prop.Value > maxTemp:
		return False
	else:
		return True	

"""Validates load case analysis be cheking that it is the right type and that it is solved"""
def validateAnalysis(result, prop):
	analysisName = prop.Value
	analysis = None
	table = prop.ParentPropertyGroup
	for ans in ExtAPI.DataModel.AnalysisList:
		if ans.Name == analysisName:
			analysis = ans
			fileName = os.path.join(analysis.WorkingDir, 'file.rst')
			if analysis.DataObjects.GetByName('Solution').PropertyValue('Status') in [0,1,2] and os.path.isfile(fileName):
				start = table.Properties["startStepColumn"]
				end = table.Properties["endStepColumn"]
				getStepsForTable(end, analysis, result)
				getStepsForTable(start, analysis, result)
				if result.Attributes['timeStamp'] and result.Attributes['Analysis_list'] and result.State == 'solved':
					if ans.Id in result.Attributes['Analysis_list']:
						temptime = os.path.getmtime(fileName)
						if result.Attributes['timeStamp'] < temptime:
							result.Suppressed = True
							ExtAPI.Application.LogWarning('"'+result.Caption+'" result in ' + result.Parent.Name + ' is outdated: "Load Case Table" contains one or more analysis which have been modified.')
							result.Attributes['timeStamp'] = None
				return True
			else:
				if result.Attributes['timeStamp'] and result.Attributes['Analysis_list'] and result.State == 'solved':
					if ans.Id in result.Attributes['Analysis_list']:
						result.Suppressed = True
						ExtAPI.Application.LogWarning('"'+result.Caption+'" result in ' + result.Parent.Name + ' is outdated: "Load Case Table" contains one or more analysis which have been modified.')
				return False

"""gets an analysis timesteps"""
def getStepsForTable(prop, analysis, result):
	prop.Options.Clear()
	if analysis.DataObjects.GetByName('Solution').PropertyValue('Status') in [0,1,2]:
		if analysis.GetResultsData() != None:
			getSteps(result, prop)
	if analysis.GetResultsData() == None:
		analysis.GetResultsData().Dispose()
		ExtAPI.GetResultsData('dummy_file.rst')		

"""checks that a value is no less than 1.0"""
def isPositiveFloat(result, prop):
	if prop.Value < 1.0:
		return False
	else:
		return True		

"""checks that an interger is postive and not a float"""
def isPositiveInteger(result, prop):
	if prop.Value > 0 and prop.Value.is_integer():
		return True
	else:
		return False	

"""checks that resultview Stress Range is only valid when load case combination is not selected"""
def loadCaseComboIsNo(result, prop):
	if prop.Value == 'Stress Range' or prop.Value == 'Strain Range':
		if  result.Properties["loadCaseCombo"].ValueString == 'No':
			return True
		else:
			return False
	else:
		return True		

#----------------------------------
#
# Onvalidate callbacks
#		
#----------------------------------

"""updates plate thickness when geoemetry is set or changed"""
def updatePlateThickness(result, prop):
	geoIds = prop.Value.Ids 
	mesh = result.Analysis.MeshData
	geometry = result.Analysis.GeoData
	bodyIds = getBodyIds_FromNodes(geoIds, mesh, geometry)
	thickness = result.Analysis.GeoData.GeoEntityById(bodyIds[0]).Thickness
	geoUnit = result.Analysis.GeoData.Assemblies[0].Unit
	userUnit = result.Analysis.CurrentConsistentUnitFromQuantityName("Length")
	thickness = units.ConvertUnit(thickness, geoUnit, userUnit, 'Length')
	prop.Value = thickness

"""updates emodule when temperature is set or changed"""
def changeEmodule(result, prop):
	table = prop.ParentPropertyGroup
	bodyIds = getBodyIds(result)
	row = table.ActiveRow
	table.Properties["emoduleColumn"].Value = findEmodule(prop.Value, result.Analysis, bodyIds[row])
	table.SaveActiveRow

"""updates emodule when temperature is set or changed"""
def changeEmodule2(result, prop):
	table = prop.ParentPropertyGroup
	try:
		geoIds = result.Properties["scope"].Properties["geometry"].Value.Ids
	except:
		pass	
	mesh = result.Analysis.MeshData
	geometry = result.Analysis.GeoData
	bodyIds = getBodyIds_FromNodes(geoIds, mesh, geometry)
	row = table.ActiveRow
	try:
		table.Properties["emoduleColumn"].Value = findEmodule(prop.Value, result.Analysis, bodyIds[row])
	except:
		pass
	table.SaveActiveRow()

"""sets display option average or unaverage"""
def setDisplayOption(result, prop):
	dispOpt = prop.ValueString
	result.ResultLocation =ResultLocationEnum.ElementNode
	if dispOpt == 'Averaged':
		result.ResultAveraging=ResultAveragingEnum.Average
	else:
		result.ResultAveraging=ResultAveragingEnum.Unaverage

#----------------------------------
#
# Onactivate Callbacks
#
#----------------------------------

"""fill bodydata input table. NB: each beam must be its own body"""
def fillLCTubularFatigueTable(result, prop):
	if prop.Value == 'Yes':
		return
	try:
		geoIds = result.Properties["scope"].Properties["geometry"].Value.Ids
	except:
		return

	table = prop.Properties["fatigueTable"]
	analysis = result.Analysis	
	geometry = analysis.GeoData
	mesh = analysis.MeshData
	geoUnit = result.Analysis.GeoData.Assemblies[0].Unit
	userUnit = result.Analysis.CurrentConsistentUnitFromQuantityName("Length")

	bodyIds = []	
	bodyNames = []
	thicknessList = []
	emodule = []
	bodyIds = getBodyIds_FromNodes(geoIds, mesh, geometry)
	userTemp = result.Analysis.PropertyValue("Temperature") 

	if len(bodyIds) == 0:
		return
	if len(bodyIds) == table.RowCount:
		return
	for i in xrange(table.RowCount):
		table.DeleteRow(i)		

	for i in xrange(len(bodyIds)):
		bodyNames.append(geometry.GeoEntityById(bodyIds[i]).Name)
		thickness = geometry.GeoEntityById(bodyIds[i]).Thickness
		thicknessList.append(units.ConvertUnit(thickness, geoUnit, userUnit, 'Length'))
		emodule.append(findEmodule(userTemp, result.Analysis, bodyIds[i]))
	for i in xrange(len(bodyIds)):
		table.AddRow()
		table.ActiveRow = i				
		bodyId = bodyIds[i]
		table.Properties["bodyNameColumn"].Value = bodyNames[i] + ' (Id = ' + str(bodyIds[i]) + ')'		
		table.Properties["plateThicknessColumn"].Value = thicknessList[i]
		table.Properties["temperatureColumn"].Value = userTemp
		table.Properties["emoduleColumn"].Value = emodule[i]
		table.SaveActiveRow()

"""fill bodydata input table. NB: each beam must be its own body"""
def fillHCTubularFatigueTable(result, prop):
	if prop.Value == 'Yes':
		return
	try:
		geoIds = result.Properties["scope"].Properties["geometry"].Value.Ids
	except:
		return

	table = prop.Properties["fatigueTable"]
	analysis = result.Analysis	
	geometry = analysis.GeoData
	mesh = analysis.MeshData
	geoUnit = result.Analysis.GeoData.Assemblies[0].Unit
	userUnit = result.Analysis.CurrentConsistentUnitFromQuantityName("Length")

	bodyIds = []	
	bodyNames = []
	thicknessList = []
	bodyIds = getBodyIds_FromNodes(geoIds, mesh, geometry)

	if len(bodyIds) == 0:
		return
	if len(bodyIds) == table.RowCount:
		return
	for i in xrange(table.RowCount):
		table.DeleteRow(i)		

	for i in xrange(len(bodyIds)):
		bodyNames.append(geometry.GeoEntityById(bodyIds[i]).Name)
		thickness = geometry.GeoEntityById(bodyIds[i]).Thickness
		thicknessList.append(units.ConvertUnit(thickness, geoUnit, userUnit, 'Length'))
	for i in xrange(len(bodyIds)):
		table.AddRow()
		table.ActiveRow = i				
		bodyId = bodyIds[i]
		table.Properties["bodyNameColumn"].Value = bodyNames[i] + ' (Id = ' + str(bodyIds[i]) + ')'		
		table.Properties["plateThicknessColumn"].Value = thicknessList[i]
		
		table.SaveActiveRow()

"""fill hot spot classification input table"""
def fillHSClassificationTable(result, prop):
	if prop.ValueString == 'No':
		return
	try:
		geoIds = result.Properties["scope"].Properties["geometry"].Value.Ids
	except:
		return

	table = prop.Properties["classificationTable"]
	analysis = result.Analysis	
	geometry = analysis.GeoData
	mesh = analysis.MeshData

	if len(geoIds) == 0:
		return

	if len(geoIds) == table.RowCount:
		return

	for i in xrange(table.RowCount):
		table.DeleteRow(i)


	for i in xrange(len(geoIds)):
		table.AddRow()
		table.ActiveRow = i
		table.Properties["hotspotIdColumn"].Value = str(geoIds[i]) 
		hsEdges = []
		hsFaces = []
		hsBodies = []
		hotspot = mesh.NodeById(geoIds[i])
		hsGeos = hotspot.GeoEntityIds
		for geo in hsGeos:
			if str(geometry.GeoEntityById(geo).Type) == "GeoEdge":
				hsEdges.append(geo)
			elif str(geometry.GeoEntityById(geo).Type) == "GeoFace":
				hsFaces.append(geo)
			elif str(geometry.GeoEntityById(geo).Type) == "GeoBody":
				hsBodies.append(geo)
				if geo not in table.Properties["braceBodyColumn"].Options:
					table.Properties["chordBodyColumn"].Options.Add(str(geo))
					table.Properties["braceBodyColumn"].Options.Add(str(geo))
		table.Properties["edgeIdColumn"].ValueString = str(hsEdges)
		table.Properties["faceIdColumn"].ValueString = str(hsFaces)
		table.Properties["bodyIdColumn"].ValueString = str(hsBodies)
		table.SaveActiveRow()							


"""fill bodydata input table"""
def fillLCPlanePlatesFatigueTable(result, prop):
	if prop.Value == 'Yes':
		return
	
	table = prop.Properties["fatigueTable"]	
	bodyIds = getBodyIds(result)
	geometry = result.Analysis.GeoData
	userTemp = result.Analysis.PropertyValue("Temperature") 
	geoUnit = result.Analysis.GeoData.Assemblies[0].Unit
	userUnit = result.Analysis.CurrentConsistentUnitFromQuantityName("Length")
	if len(bodyIds) == 0:
		return

	if len(bodyIds) == table.RowCount:
		return

	for i in xrange(table.RowCount):
		table.DeleteRow(i)

	bodyNames = []
	emodule = []
	thicknessList = []
	for i in xrange(len(bodyIds)):
		bodyNames.append(geometry.GeoEntityById(bodyIds[i]).Name)
		emodule.append(findEmodule(userTemp, result.Analysis, bodyIds[i]))
		thickness = result.Analysis.GeoData.GeoEntityById(bodyIds[i]).Thickness
		thicknessList.append(units.ConvertUnit(thickness, geoUnit, userUnit, 'Length'))

	for i in xrange(len(bodyIds)):
		table.AddRow()
		table.ActiveRow = i
		bodyId = bodyIds[i]
		table.Properties["bodyNameColumn"].Value = bodyNames[i] + ' (Id = ' + str(bodyIds[i]) + ')'
		table.Properties["temperatureColumn"].Value = userTemp
		table.Properties["emoduleColumn"].Value = emodule[i]
		table.Properties["plateThicknessColumn"].Value = thicknessList[i]
		table.SaveActiveRow()

"""fill bodydata input table"""
def fillLCScreeningFatigueTable(result, prop):
	if prop.Value == 'Yes':
		return
	
	table = prop.Properties["fatigueTable"]	
	bodyIds = getBodyIds(result)
	geometry = result.Analysis.GeoData
	userTemp = result.Analysis.PropertyValue("Temperature") 

	if len(bodyIds) == 0:
		return

	if len(bodyIds) == table.RowCount:
		return

	for i in xrange(table.RowCount):
		table.DeleteRow(i)

	bodyNames = []
	emodule = []
	for i in xrange(len(bodyIds)):
		bodyNames.append(geometry.GeoEntityById(bodyIds[i]).Name)
		emodule.append(findEmodule(userTemp, result.Analysis, bodyIds[i]))

	for i in xrange(len(bodyIds)):
		table.AddRow()
		table.ActiveRow = i
		bodyId = bodyIds[i]
		table.Properties["bodyNameColumn"].Value = bodyNames[i] + ' (Id = ' + str(bodyIds[i]) + ')'
		table.Properties["temperatureColumn"].Value = userTemp
		table.Properties["emoduleColumn"].Value = emodule[i]
		table.SaveActiveRow()

"""fill bodydata input table"""
def fillHCPlanePlatesFatigueTable(result, prop):
	if prop.Value == 'Yes':
		return

	table = prop.Properties["fatigueTable"]	
	bodyIds = getBodyIds(result)
	geometry = result.Analysis.GeoData
	geoUnit = geometry.Assemblies[0].Unit
	userUnit = result.Analysis.CurrentConsistentUnitFromQuantityName("Length")
	refThick = (units.ConvertUnit(25, 'mm', userUnit, 'Length'))

	if len(bodyIds) == 0:
		return

	if len(bodyIds) == table.RowCount:
		return

	for i in xrange(table.RowCount):
		table.DeleteRow(i)

	bodyNames = []
	thicknessList = []
	for i in xrange(len(bodyIds)):
		bodyNames.append(geometry.GeoEntityById(bodyIds[i]).Name)
		thickness = result.Analysis.GeoData.GeoEntityById(bodyIds[i]).Thickness
		thicknessList.append(units.ConvertUnit(thickness, geoUnit, userUnit, 'Length'))

	for i in xrange(len(bodyIds)):
		table.AddRow()
		table.ActiveRow = i
		bodyId = bodyIds[i]
		table.Properties["bodyNameColumn"].Value = bodyNames[i] + ' (Id = ' + str(bodyIds[i]) + ')'
		table.Properties["plateThicknessColumn"].Value = thicknessList[i]
		table.Properties["referenceThicknessColumn"].Value = refThick
		table.SaveActiveRow()

"""fill bodydata input table"""
def fillHCScreeningFatigueTable(result, prop):
	if prop.Value == 'Yes':
		return

	table = prop.Properties["fatigueTable"]	
	bodyIds = getBodyIds(result)
	geometry = result.Analysis.GeoData

	if len(bodyIds) == 0:
		return

	if len(bodyIds) == table.RowCount:
		return

	for i in xrange(table.RowCount):
		table.DeleteRow(i)

	bodyNames = []
	for i in xrange(len(bodyIds)):
		bodyNames.append(geometry.GeoEntityById(bodyIds[i]).Name)

	for i in xrange(len(bodyIds)):
		table.AddRow()
		table.ActiveRow = i
		bodyId = bodyIds[i]
		table.Properties["bodyNameColumn"].Value = bodyNames[i] + ' (Id = ' + str(bodyIds[i]) + ')'
		table.SaveActiveRow()

"""collects all available analyses for load case and fills the option in load case table"""
def getAnalyses(result, prop):
	prop.Options.Clear()
	for analysis in ExtAPI.DataModel.AnalysisList:
		if analysis.PhysicsType == 'Mechanical':
			prop.Options.Add(analysis.Name)		

"""gets timesteps for an analysis"""
def getSteps(result, prop):
	try:
		prop.Options.Clear()
		if prop.Name == 'startStep' or prop.Name == 'startStepColumn':
			prop.Options.Add('0')
		for step in xrange(1,len(result.Analysis.StepsEndTime)+1):
			prop.Options.Add(str(step))			
	except:
		prop.Options.Clear()

#----------------------------------
#
# Isvisible callbacks
#
#----------------------------------

"""checks if geometry is chosen"""
def geometryIsChosen(result, prop):
	try:
		geoIds = result.Properties["scope"].Properties["geometry"].Value.Ids
	except:	
		geoIds = []
	try:
		if len(geoIds) == 0:
			return False
		else: 
			return True	
	except:
		return False	

"""checks if property high temperature is set to yes"""
def isHighTemp(result, prop):
	highTemp = result.Properties["snClass"].Properties["snEqual"].Properties["highTemp"].ValueString
	if highTemp == 'Yes':
		return True
	else:
		return False

"""checks if damage is selected and no load case combination"""
def damageAndNoLoadCaseCombo(result, prop):
	if result.Properties["resultView"].ValueString == 'Damage':
		if result.Properties["loadCaseCombo"].ValueString == 'No':
			return True
		else:
			return False
	else:
		return False	

"""checks if damage and  load case combination is selected"""
def damageAndLoadCaseCombo(result, prop):
	if result.Properties["resultView"].ValueString == 'Damage':
		if result.Properties["loadCaseCombo"].ValueString == 'Yes':
			return True
		else:
			return False
	else:
		return False		

#----------------------------------
#
# Oninit callbacks
#
#----------------------------------

"""sets the plate thickness at result object initiation"""
def initializePlateThickness(result, prop):
	bodyIds = getBodyIds(result)
	thickness = result.Analysis.GeoData.GeoEntityById(bodyIds[0]).Thickness
	geoUnit = result.Analysis.GeoData.Assemblies[0].Unit
	userUnit = result.Analysis.CurrentConsistentUnitFromQuantityName("Length")
	thickness = units.ConvertUnit(thickness, geoUnit, userUnit, 'Length')
	prop.Value = thickness

"""sets the temperature at result object initiation"""
def initializeTemperature(result, prop):
	prop.Value = result.Analysis.PropertyValue("Temperature") 

"""sets the load case table size when table is clicked"""
def initializeLoadCaseTable(result, prop):
	prop.Controller.size = [800,300]
