#\file hcScreening.py
#
#----------------------------------------------------------------------------------
#
#   Implemented controller for high cycle fatigue calculation in base material
#   
#----------------------------------------------------------------------------------
#----------------------------------------------------------------------------------
#
# author: Tone Wermundsen
# Last update: 18.12.2014
# ansys v15 ACT 
#
#-----------------------------------------------------------------------------------

#----------------------------------------------------------------------------------- 
#   Nested controller class that handles all evaluation behavior of the result object
#-----------------------------------------------------------------------------------

class HCScreeningController():
	def __init__(self, ExtAPI, result):
		self.ExtAPI = ExtAPI
		self.result = result

		displayOption = result.Properties["integrationPointResult"].Properties["displayOption"]
		setDisplayOption(result, displayOption)

	def evaluate(self, result, stepInfo, collector):
		result.Attributes['timeStamp'] = 0
		result.Attributes['Analysis_list'] = []
		ExtAPI.Log.WriteMessage("ready for HC Screening")

		#decs
		myAnalysis = result.Analysis 
		myMesh = myAnalysis.MeshData
		myGeometry = myAnalysis.GeoData
		myReader = myAnalysis.GetResultsData()

		#units
		meshUnit = myMesh.Unit 
		geoUnit = myGeometry.Assemblies[0].Unit
		userUnitTemp = myAnalysis.CurrentConsistentUnitFromQuantityName("Temperature")
		userUnitStress = myAnalysis.CurrentConsistentUnitFromQuantityName("Stress")
		userUnitLength = myAnalysis.CurrentConsistentUnitFromQuantityName("Length")

		#user input
		geoIds = result.Properties["scope"].Properties["geometry"].Value.Ids
		shellOption = result.Properties["scope"].Properties["shellOption"].ValueString
		resultView = result.Properties["resultView"].ValueString
		displayOption = result.Properties["integrationPointResult"].Properties["displayOption"].ValueString
		lcCombo = result.Properties["loadCaseCombo"].ValueString
		snEqual = result.Properties["snClass"].Properties["snEqual"].ValueString
		highTemp = result.Properties["snClass"].Properties["snEqual"].Properties["highTemp"].ValueString
		lcTable = result.Properties["loadCaseCombo"].Properties["loadCaseTable"]
		stressOption = result.Properties["stressOption"].ValueString


		bodyIds = getBodyIds(result)
		if len(bodyIds) == 0:
			MessageBox.Show(ExtAPI.UserInterface.MainWindow, "Unvalid input.", 'Error', MessageBoxType.Error, MessageBoxButtons.OK) 
			return
		bodyData = fillBodyDataHCScreening(bodyIds, result, snEqual, myGeometry, myAnalysis, userUnitLength, meshUnit, highTemp, userUnitTemp)	
		loadCase = fillLoadCaseInfo(lcCombo, myAnalysis, myReader, result, resultView, lcTable)
		summedLogN = {}; summedDamage = {}; summedN = {}
		for elementId in collector.Ids:
			summedLogN[elementId] = []
			summedDamage[elementId] = []
			summedN[elementId] = []

		excepter = 0	
		for lc in loadCase:
			loadCaseAnalysis = loadCase[lc][1]
			startTime = loadCase[lc][2]
			endTime = loadCase[lc][3]
			loadCaseReader = loadCaseAnalysis.GetResultsData()
			loadCaseResultItemStress = loadCaseReader.GetResult('S')
			resultUnitStress = loadCaseResultItemStress.GetComponentInfo('X').Unit 
			convUnitStress = units.ConvertUnit(1., resultUnitStress, 'MPa', 'Stress') 
			numberOfCycles = loadCase[lc][4]
			Shell = False

			resultComponentsT1 = {}; resultComponentsT2 = {}
			loadCaseReader.CurrentResultSet = loadCaseReader.ListTimeFreq.IndexOf(endTime)+1
			for elementId in collector.Ids:
				elementType = myMesh.ElementById(elementId).Type 
				if str(elementType) in ['kQuad8','kQuad4','kTri3','kTri6']:
					Shell = True
				else:
					Shell = False	
				resultComponentsT2[elementId] = getAllResultComponents_FatigueDNV(loadCaseResultItemStress, elementId, convUnitStress, shellOption, Shell)
			if startTime != 0:
				loadCaseReader.CurrentResultSet = loadCaseReader.ListTimeFreq.IndexOf(startTime)+1
				for elementId in collector.Ids:
					elementType = myMesh.ElementById(elementId).Type			
					if str(elementType) in ['kQuad8','kQuad4','kTri3','kTri6']:
						Shell = True 
					else:	
						Shell = False
					resultComponentsT1[elementId] = getAllResultComponents_FatigueDNV(loadCaseResultItemStress, elementId, convUnitStress, shellOption, Shell)
			else:
				for elementId in collector.Ids:
					resultComponentsT1[elementId] = [0.0]*len(resultComponentsT2[elementId])

			resultSet = {}		
			for elementId in collector.Ids:
				resultSet[elementId] = calculatePrincipal_FatigueDNV(resultComponentsT1[elementId], resultComponentsT2[elementId], True, stressOption)		

			if resultView != 'Stress Range':
				logN = {}; damage = {}; N = {}
				logN = calculateLifeHighCycleNormal(myMesh, collector, resultSet, bodyData)


				for elementId in collector.Ids:
					N[elementId] = map(operator.pow, len(logN[elementId])*(10, ), logN[elementId])
					if resultView == 'Damage':
						damage[elementId] = calculateDamage(N[elementId], numberOfCycles)

				#check if there are any N smaller than 1e4
				for elementId in collector.Ids:
					for value in N[elementId]:
						if value < 1e4:
							excepter += 1
							break

				for elementId in collector.Ids:
					if lcCombo == 'No':
						summedN[elementId] = N[elementId]
						summedLogN[elementId] = logN[elementId]
						if resultView == 'Damage':
							summedDamage[elementId] = damage[elementId]
					else:
						if lc == 0:	
							summedN[elementId] = map(operator.div, len(N[elementId])*(1, ), N[elementId])
							if resultView == 'Damage':
								summedDamage[elementId] = damage[elementId]
						if lc != 0:
							for i in xrange(len(N[elementId])):
								try:
									summedN[elementId][i] += 1/N[elementId][i]
								except:
									pass
								if resultView == 'Damage':
									summedDamage[elementId][i] += damage[elementId][i]
						if lc == lcTable.RowCount-1:
							summedN[elementId] = map(operator.div, len(summedN[elementId])*(1, ), summedN[elementId])
							summedLogN[elementId] = map(math.log10, summedN[elementId])

		if resultView == 'Stress Range':
			for elementId in collector.Ids:
				collector.SetValues(elementId, resultSet[elementId])
		if resultView == 'Damage':
			for elementId in collector.Ids:
				collector.SetValues(elementId, summedDamage[elementId])
		if resultView == 'Cycles To Failure':
			for elementId in collector.Ids:
				collector.SetValues(elementId, summedN[elementId])	
		if resultView == 'Logarithm Of Cycles To Failure To Base 10':
			for elementId in collector.Ids:
				collector.SetValues(elementId, summedLogN[elementId])																					

		result.Attributes['timeStamp'] = time.time()
		
#----------------------------------------------------------------------------------- 
#   Fills bodyData with information about each bodies environment, snCurve properties and material properties
#-----------------------------------------------------------------------------------
def fillBodyDataHCScreening(bodyIds, result, snEqual, geometry, analysis, userUnitLength, meshUnit, highTemp, userUnitTemp):
	if highTemp == 'Yes':
		temperature = result.Properties["snClass"].Properties["snEqual"].Properties["temperature"].Value
		temperature = units.ConvertUnit(temperature, userUnitTemp, 'C', 'Temperature')
	bodyData = {}
	if snEqual == 'Yes':
		environment = result.Properties["snClass"].Properties["snEqual"].Properties["environment"].ValueString
		snCurve = result.Properties["snClass"].Properties["snEqual"].Properties["snCurve"].ValueString	
		thicknessModFac = result.Properties["snClass"].Properties["snEqual"].Properties["thicknessModFac"].Value
		m1, logA1, m2, logA2, stressLimit, k = SNData_FatigueDNV(environment, snCurve)
		if highTemp == 'Yes':
			RT = 1.0376 - 0.239*10**(-3)*temp - 1.372*10**(-6)*(temp**2)
			logA1RT = logA1 + m1*math.log10(RT)
			logA1 = logA1RT
			RT = 1.0376 - 0.239*10**(-3)*temp - 1.372*10**(-6)*(temp**2)
			logA2RT = logA2 + m1*math.log10(RT)
			logA2 = logA2RT
		if thicknessModFac < 1:
			thicknessModFac = 1
		if snCurve == 'HighStrengthSteel' or environment == 'Seawater Free Corrosion':
			crackLimit =  0
		else:		
			limitPart1 = 10**(logA2 - logA1)
			limitPart2 = 1/(m2-m1)
			crackLimit = limitPart1**limitPart2	
		appliedThickModFac = thicknessModFac**k
		for bodyId in bodyIds:
			bodyData[bodyId] = [snCurve, environment, m1, logA1, m2, logA2, stressLimit, k, appliedThickModFac, crackLimit]
	else:
		table = result.Properties["snClass"].Properties["snEqual"].Properties["fatigueTable"]
		for row in xrange(table.RowCount):
			table.ActiveRow = row
			bodyId = bodyIds[row]
			snCurve = table.Properties["curveColumn"].ValueString
			environment = table.Properties["environmentColumn"].ValueString
			thicknessModFac = table.Properties["thicknessModFacColumn"].Value
			m1, logA1, m2, logA2, stressLimit, k = SNData_FatigueDNV(environment, snCurve)
			if thicknessModFac < 1:	
				thicknessModFac = 1
			if highTemp == 'Yes':	
				RT = 1.0376 - 0.239*10**(-3)*temp - 1.372*10**(-6)*(temp**2)
				logA1RT = logA1 + m1*math.log10(RT)
				logA1 = logA1RT
				RT = 1.0376 - 0.239*10**(-3)*temp - 1.372*10**(-6)*(temp**2)
				logA2RT = logA2 + m1*math.log10(RT)
				logA2 = logA2RT
			if snCurve == 'HighStrengthSteel' or environment == 'Seawater Free Corrosion':
				crackLimit =  0
			else:	
				limitPart1 = 10**(logA2 - logA1)
				limitPart2 = 1/(m2-m1)
				crackLimit = limitPart1**limitPart2
			appliedThickModFac = thicknessModFac**k
			bodyData[bodyId] = [snCurve, environment, m1, logA1, m2, logA2, stressLimit, k, appliedThickModFac, crackLimit]
	return bodyData		