#\file hcTubular.py
#
#----------------------------------------------------------------------------------
#
#   Implemented controller for high cycle fatigue calculation on tubular joint 
#   
#----------------------------------------------------------------------------------
#----------------------------------------------------------------------------------
#
# author: Tone Wermundsen
# Last update: 18.12.2014
# ansys v15 ACT 
#
#-----------------------------------------------------------------------------------

#----------------------------------------------------------------------------------- 
#   Nested controller class that handles all evaluation behavior of the result object
#-----------------------------------------------------------------------------------
class HCTubularController():
	def __init__(self, ExtAPI, result):
		self.ExtAPI = ExtAPI
		self.result = result

	def evaluate(self, result, stepInfo, collector):
		result.Attributes['timeStamp'] = 0
		result.Attributes['Analysis_list'] = []
		ExtAPI.Log.WriteMessage("ready for HC tubular")
		MessageBox.Show(ExtAPI.UserInterface.MainWindow, "Unfinished result object, result may not apply or be viewed at all", 'Warning', MessageBoxType.Warning, MessageBoxButtons.OK) 
		
		#decs
		myAnalysis = result.Analysis 
		myMesh = myAnalysis.MeshData
		myGeometry = myAnalysis.GeoData
		myReader = myAnalysis.GetResultsData()

		#units
		meshUnit = myMesh.Unit 
		geoUnit = myGeometry.Assemblies[0].Unit
		userUnitTemp = myAnalysis.CurrentConsistentUnitFromQuantityName("Temperature")
		userUnitStress = myAnalysis.CurrentConsistentUnitFromQuantityName("Stress")
		userUnitLength = myAnalysis.CurrentConsistentUnitFromQuantityName("Length")

		#user input
		geoSelected = result.Properties["hotspotClassification"].Properties["geoSelected"].ValueString
		if geoSelected == 'No':
			msg = "Please select geometry and fill in classification table after selecting 'yes' at property Geometry is Selected"
			MessageBox.Show(ExtAPI.UserInterface.MainWindow, msg, 'Error', MessageBoxType.Error, MessageBoxButtons.OK)
			return
		classificationTable = result.Properties["hotspotClassification"].Properties["geoSelected"].Properties["classificationTable"]
		geoIds = result.Properties["scope"].Properties["geometry"].Value.Ids
		shellOption = result.Properties["scope"].Properties["shellOption"].ValueString
		resultView = result.Properties["resultView"].ValueString
		displayOption = result.Properties["integrationPointResult"].Properties["displayOption"].ValueString
		lcCombo = result.Properties["loadCaseCombo"].ValueString
		snEqual = result.Properties["snClass"].Properties["snEqual"].ValueString
		highTemp = result.Properties["snClass"].Properties["snEqual"].Properties["highTemp"].ValueString
		lcTable = result.Properties["loadCaseCombo"].Properties["loadCaseTable"]
		stressOption = result.Properties["stressOption"].ValueString

		bodyIds = getBodyIds_FromNodes(geoIds, myMesh, myGeometry)
		if len(bodyIds) == 0:
			MessageBox.Show(ExtAPI.UserInterface.MainWindow, "Unvalid input.", 'Error', MessageBoxType.Error, MessageBoxButtons.OK) 
			return
		bodyData = fillBodyDataHCTubular(bodyIds, result, snEqual, myGeometry, myAnalysis, userUnitLength, meshUnit, highTemp, userUnitTemp, geoUnit)	
		myHotspots = geoIds
		if len(myHotspots) == 0:
			MessageBox.Show(ExtAPI.UserInterface.MainWindow, "Could not recognize any hot spots. Try changing geometry", 'Warning', MessageBoxType.Warning, MessageBoxButtons.OK)
			return
		hotspotData = fillHotSpotData(myHotspots, result, classificationTable, myMesh, myGeometry)	
		loadCase = fillLoadCaseInfo(lcCombo, myAnalysis, myReader, result, resultView, lcTable)

		smallExtLocations = {}; smallExtElements = {}; largeExtLocations = {}; largeExtElements = {}
		for hotspotId in hotspotData:
			hotspot = myMesh.NodeById(hotspotId)
			hotspotLocation = [hotspot.X, hotspot.Y, hotspot.Z]
			smallExtLocations[hotspotId] = getTubularLocations('small', hotspotId, hotspotData[hotspotId], bodyData, myMesh, myGeometry)
			largeExtLocations[hotspotId] = getTubularLocations('large', hotspotId, hotspotData[hotspotId], bodyData, myMesh, myGeometry)
			smallExtElements[hotspotId] = getTubularElements(myMesh, hotspotId, smallExtLocations[hotspotId], bodyData, myGeometry, geoUnit, meshUnit)
		
		result.Attributes['timeStamp'] = time.time()
	
#----------------------------------------------------------------------------------- 
#   Fills bodyData with information about each bodies environment, snCurve properties and material properties
#-----------------------------------------------------------------------------------
def fillBodyDataHCTubular(bodyIds, result, snEqual, geometry, analysis, userUnitLength, meshUnit, highTemp, userUnitTemp, geoUnit):
	if highTemp == 'Yes':
		temperature = result.Properties["snClass"].Properties["snEqual"].Properties["temperature"].Value
		temperature = units.ConvertUnit(temperature, userUnitTemp, 'C', 'Temperature')
	bodyData = {}
	if snEqual == 'Yes':
		environment = result.Properties["snClass"].Properties["snEqual"].Properties["environment"].ValueString
		snCurve = result.Properties["snClass"].Properties["snEqual"].Properties["snCurve"].ValueString	
		plateThickness = result.Properties["snClass"].Properties["snEqual"].Properties["plateThickness"].Value
		referenceThickness = result.Properties["snClass"].Properties["snEqual"].Properties["referenceThickness"].Value
		thicknessModFac = plateThickness/referenceThickness
		m1, logA1, m2, logA2, stressLimit, k = SNData_FatigueDNV(environment, snCurve)
		if highTemp == 'Yes':
			RT = 1.0376 - 0.239*10**(-3)*temp - 1.372*10**(-6)*(temp**2)
			logA1RT = logA1 + m1*math.log10(RT)
			logA1 = logA1RT
			RT = 1.0376 - 0.239*10**(-3)*temp - 1.372*10**(-6)*(temp**2)
			logA2RT = logA2 + m1*math.log10(RT)
			logA2 = logA2RT
		if thicknessModFac < 1:
			thicknessModFac = 1
		if snCurve == 'HighStrengthSteel' or environment == 'Seawater Free Corrosion':
			crackLimit = 0	
		else:	
			limitPart1 = 10**(logA2 - logA1)
			limitPart2 = 1/(m2-m1)
			crackLimit = limitPart1**limitPart2	
		appliedThickModFac = thicknessModFac**k
		plateThickness = units.ConvertUnit(plateThickness, userUnitLength, meshUnit, "Length")
		for bodyId in bodyIds:
			bodyCentroid = geometry.GeoEntityById(bodyId).Centroid 
			edges = geometry.GeoEntityById(bodyId).Edges 
			for edge in edges:
				edgeId = edge.Id 
				if str(geometry.GeoEntityById(edgeId).CurveType) == 'GeoCurveCircle':
					circumference = geometry.GeoEntityById(edgeId).Length 
					circumference = units.ConvertUnit(circumference, geoUnit, meshUnit, 'Length')
					edgeCentroid = geometry.GeoEntityById(edgeId).Centroid
					pipeDirection = [edgeCentroid[0]-bodyCentroid[0], edgeCentroid[1]-bodyCentroid[1], edgeCentroid[2]-bodyCentroid[2]]
					pipeDirection = unitVector(pipeDirection)
					break
			radius = circumference/(2*math.pi)		
			bodyData[bodyId] = [snCurve, environment, m1, logA1, m2, logA2, stressLimit, k, appliedThickModFac, crackLimit, plateThickness, radius, pipeDirection, circumference]
	else:
		table = result.Properties["snClass"].Properties["snEqual"].Properties["fatigueTable"]
		for row in xrange(table.RowCount):
			table.ActiveRow = row
			bodyId = bodyIds[row]
			snCurve = table.Properties["curveColumn"].ValueString
			environment = table.Properties["environmentColumn"].ValueString
			plateThickness = table.Properties["plateThicknessColumn"].Value
			referenceThickness = table.Properties["referenceThicknessColumn"].Value
			thicknessModFac = plateThickness/referenceThickness
			m1, logA1, m2, logA2, stressLimit, k = SNData_FatigueDNV(environment, snCurve)
			if thicknessModFac < 1:	
				thicknessModFac = 1
			if highTemp == 'Yes':	
				RT = 1.0376 - 0.239*10**(-3)*temp - 1.372*10**(-6)*(temp**2)
				logA1RT = logA1 + m1*math.log10(RT)
				logA1 = logA1RT
				RT = 1.0376 - 0.239*10**(-3)*temp - 1.372*10**(-6)*(temp**2)
				logA2RT = logA2 + m1*math.log10(RT)
				logA2 = logA2RT
			if snCurve == 'HighStrengthSteel' or environment == 'Seawater Free Corrosion':
				crackLimit = 0
			else:		
				limitPart1 = 10**(logA2 - logA1)
				limitPart2 = 1/(m2-m1)
				crackLimit = limitPart1**limitPart2
			appliedThickModFac = thicknessModFac**k
			plateThickness = units.ConvertUnit(plateThickness, userUnitLength, meshUnit, "Length")

			bodyCentroid = geometry.GeoEntityById(bodyId).Centroid
			edges = geometry.GeoEntityById(bodyId).Edges
			for edge in edges:
				edgeId = edge.Id 
				if str(geometry.GeoEntityById(edgeId).CurveType) == 'GeoCurveCircle':
					circumference = geometry.GeoEntityById(edgeId).Length 
					circumference = units.ConvertUnit(circumference, geoUnit, meshUnit, 'Length')
					edgeCentroid = geometry.GeoEntityById(edgeId).Centroid
					pipeDirection = [edgeCentroid[0]-bodyCentroid[0], edgeCentroid[1]-bodyCentroid[1], edgeCentroid[2]-bodyCentroid[2]]
					pipeDirection = unitVector(pipeDirection)
					break
				radius = circumference/(2*math.pi)	
			bodyData[bodyId] = [snCurve, environment, m1, logA1, m2, logA2, stressLimit, k, appliedThickModFac, crackLimit, plateThickness, radius, pipeDirection, circumference]
	return bodyData	