#\file lcPlanePlates.py
#
#----------------------------------------------------------------------------------
#
#   Implemented controller for low cycle fatigue calculation on plane plates welds
#   
#----------------------------------------------------------------------------------
#----------------------------------------------------------------------------------
#
# author: Tone Wermundsen
# Last update: 18.12.2014
# ansys v15 ACT 
#
#-----------------------------------------------------------------------------------

#----------------------------------------------------------------------------------- 
#   Nested controller class that handles all evaluation behavior of the result object
#-----------------------------------------------------------------------------------
class LCPlanePlatesController():
	def __init__(self, ExtAPI, result):
		self.ExtAPI = ExtAPI
		self.result = result

	def evaluate(self, result, stepInfo, collector):
		result.Attributes['timeStamp'] = 0
		result.Attributes['Analysis_list'] = []
		ExtAPI.Log.WriteMessage("ready for LC PlanePlates")

		#decs
		myAnalysis = result.Analysis 
		myMesh = myAnalysis.MeshData
		myGeometry = myAnalysis.GeoData
		myReader = myAnalysis.GetResultsData()

		#units
		meshUnit = myMesh.Unit 
		geoUnit = myGeometry.Assemblies[0].Unit
		userUnitTemp = myAnalysis.CurrentConsistentUnitFromQuantityName("Temperature")
		userUnitStress = myAnalysis.CurrentConsistentUnitFromQuantityName("Stress")
		userUnitLength = myAnalysis.CurrentConsistentUnitFromQuantityName("Length")

		#user input
		geoIds = result.Properties["scope"].Properties["geometry"].Value.Ids
		shellOption = result.Properties["scope"].Properties["shellOption"].ValueString
		resultView = result.Properties["resultView"].ValueString
		#displayOption = result.Properties["integrationPointResult"].Properties["displayOption"].ValueString
		lcCombo = result.Properties["loadCaseCombo"].ValueString
		snEqual = result.Properties["snClass"].Properties["snEqual"].ValueString
		N0 = result.Properties["Advanced"].Properties["initialGuess"].Value
		lcTable = result.Properties["loadCaseCombo"].Properties["loadCaseTable"]

		bodyIds = getBodyIds(result)
		if len(bodyIds) == 0:
			MessageBox.Show(ExtAPI.UserInterface.MainWindow, "Unvalid input.", 'Error', MessageBoxType.Error, MessageBoxButtons.OK) 
			return
		else:
			myWelds 	= findWelds(geoIds, myGeometry)
			if len(myWelds) == 0:
				MessageBox.Show(ExtAPI.UserInterface.MainWindow, "Could not recognize any welds. Try changing geometry to type edge", 'Warning', MessageBoxType.Warning, MessageBoxButtons.OK)
			myHotspots  = findHotspots(myWelds, myMesh)
			myFaces 	= findFacesConnectedToEdge(myWelds, myGeometry)
			allHotspots = []
			for edgeId in myWelds:
				for hotspotId in myHotspots[edgeId]:
					if hotspotId not in allHotspots:
						allHotspots.append(hotspotId)	
		bodyData = fillBodyDataLCPlanePlates(bodyIds, result, snEqual, myGeometry, myAnalysis, userUnitLength, meshUnit)

		loadCase = fillLoadCaseInfo(lcCombo, myAnalysis, myReader, result, resultView, lcTable)

		#find ext
		extDirections = {}; smallExtLocations = {}; largeExtLocations = {}; smallExtElements = {}; largeExtElements = {}
		for edgeId in myWelds:
			for hotspotId in myHotspots[edgeId]:
				hotspot1 	 = myMesh.NodeById(myHotspots[edgeId][0])
				hotspot2 	 = myMesh.NodeById(myHotspots[edgeId][1])
				hotspot  	 = myMesh.NodeById(hotspotId)
				
				if hotspotId == myHotspots[edgeId][0]:
					weldLine = [hotspot2.X - hotspot1.X, hotspot2.Y - hotspot1.Y, hotspot2.Z - hotspot1.Z]
				else:
					weldLine = [hotspot1.X - hotspot2.X, hotspot1.Y - hotspot2.Y, hotspot1.Z - hotspot2.Z]
				middlePoint  = [(hotspot2.X + hotspot1.X)/2, (hotspot2.Y + hotspot1.Y)/2, (hotspot2.Z + hotspot1.Z)/2]	
				for faceId in myFaces[edgeId]:
					extDirections[edgeId, faceId, hotspotId]     = getExtDirections(myMesh, myGeometry, hotspotId, faceId, edgeId, weldLine)
					faceBody 							   		 = myGeometry.GeoEntityById(faceId).Bodies[0].Id
					faceThickness 						   		 = bodyData[faceBody][4]
					smallExtLocations[edgeId, faceId, hotspotId] = getExtCoordinates(myMesh, hotspotId, extDirections[edgeId, faceId, hotspotId], 0.5*faceThickness)	
					largeExtLocations[edgeId, faceId, hotspotId] = getExtCoordinates(myMesh, hotspotId, extDirections[edgeId, faceId, hotspotId], 1.5*faceThickness)
					smallExtElements[edgeId, faceId, hotspotId]  = getExtElements(myMesh, hotspotId, smallExtLocations[edgeId, faceId, hotspotId], faceId, middlePoint)
					largeExtElements[edgeId, faceId, hotspotId]  = getExtElements(myMesh, hotspotId, largeExtLocations[edgeId, faceId, hotspotId], faceId, middlePoint)

		#ok! So now you have all ext locs. Some are outside the models boundary, and some are within. To extrapolate, make sure both small and large is within
		eliminateOutOfBoundElements(smallExtElements, largeExtElements, myWelds, myHotspots, myFaces) 			

		extElements = getAllExtrapolationElements(smallExtElements, largeExtElements)	
		summedN = {}; summedLogN = {}; summedDamage = {}
		for Id in (set(collector.Ids) & set(allHotspots)):
			summedN[Id] = []
			summedLogN[Id] = []
			summedDamage[Id] = []

		for lc in loadCase:
			lcAnalysis = loadCase[lc][1]
			lcReader = lcAnalysis.GetResultsData()
			lcResultItem = lcReader.GetResult('EPTO')
			lcEndTime = loadCase[lc][3]
			lcStartTime = loadCase[lc][2]
			numberOfCycles = loadCase[lc][4]

			nodeComponentsT1 = {}; nodeComponentsT2 = {}
			lcReader.CurrentResultSet = lcReader.ListTimeFreq.IndexOf(lcEndTime)+1
			nodeComponentsT2 = gatherNodeComponents(myMesh, extElements, lcResultItem, lcEndTime, lc)

			if lcStartTime != 0:
				lcReader.CurrentResultSet = lcReader.ListTimeFreq.IndexOf(lcStartTime)+1
				nodeComponentsT1 = gatherNodeComponents(myMesh, extElements, lcResultItem, lcStartTime, lc)
			else:
				for key in nodeComponentsT2:
					nodeComponentsT1[key] = [0.0]*len(nodeComponentsT2[key])	
			
			nodeValues = getNodeValues(nodeComponentsT1, nodeComponentsT2, False, 'nothing')	

			smallExtValues = {}; largeExtValues = {}; extValues = {}; hotSpotValue = {}
			for edgeId in myWelds:
				for hotspotId in myHotspots[edgeId]:
					for faceId in myFaces[edgeId]:
						key = (edgeId, faceId, hotspotId)
						faceBody  = myGeometry.GeoEntityById(faceId).Bodies[0].Id
						faceThickness = bodyData[faceBody][4]
						smallExtValues[key] = findInterpolatedValue(myMesh, smallExtElements[key], smallExtLocations[key], nodeValues, loadCase[lc][3], lc)
						largeExtValues[key] = findInterpolatedValue(myMesh, largeExtElements[key], largeExtLocations[key], nodeValues, loadCase[lc][3], lc)
						extValues[key] = getCombinedExtValues(smallExtValues[key], largeExtValues[key], faceThickness)

			for hsId in allHotspots:
				for edgeId in myWelds:
					if hsId in myHotspots[edgeId]:
						hotSpotValue[hsId] = [0,0,0]
						for faceId in myFaces[edgeId]:
							key = (edgeId, faceId, hsId)
							for i in xrange(len(extValues[key])):
								if max(extValues[key][i]) > max(hotSpotValue[hsId]):
									hotSpotValue[hsId] = extValues[key][i]	

			if resultView != 'Strain Range':
				logN = {}; N = {}; damage = {}
				N, excepter, highLife, errorLog = calculateLowCyclePlanePlates(myMesh, hotSpotValue, bodyData, N0)
				for Id in N:
					logN[Id] = map(math.log10, N[Id])
					if resultView == 'Damage':
						damage[Id] = calculateDamage(N[Id], numberOfCycles)

				
	

				for Id in N:
					if lcCombo == 'No':
						summedN[Id] = N[Id]
						summedLogN[Id] = logN[Id]
						if resultView == 'Damage':
							summedDamage[Id] = damage[Id]
					else:
						if lc == 0:
							summedN[Id] = map(operator.div, len(N[Id])*(1, ), N[Id])
							if resultView == 'Damage':
								summedDamage[Id] = damage[Id]		
						if lc != 0:
							for i in xrange(len(N[Id])):
								try:
									summedN[Id][i] += 1/N[Id][i]
								except:
									pass 
								if resultView == 'Damage':
									summedDamage[Id][i] += damage[Id][i]
						if lc == lcTable.RowCount-1:
							summedN[Id] = map(operator.div, len(summedN[Id])*(1, ), summedN[Id])
							summedLogN[Id] = map(math.log10, summedN[Id])	

		if shellOption == 'Top':
			i = 1
		if shellOption == 'Middle':
			i = 2
		else:
			i = 0
		if resultView == 'Strain Range':	
			for Id in allHotspots:		
				collector.SetValues(Id, [hotSpotValue[Id][i]])
		elif resultView == 'Damage':
			for Id in (set(collector.Ids) & set(allHotspots)):
				collector.SetValues(Id, [summedDamage[Id][i]])
		elif resultView == 'Cycles To Failure':
			for Id in (set(collector.Ids) & set(allHotspots)):
				collector.SetValues(Id, [summedN[Id][i]])
		elif resultView == 'Logarithm Of Cycles To Failure To Base 10':
			for Id in (set(collector.Ids) & set(allHotspots)):
				collector.SetValues(Id, [summedLogN[Id][i]])									
	


		result.Attributes['timeStamp'] = time.time()

#----------------------------------------------------------------------------------- 
#   Fills bodyData with information about each bodies environment and material properties
#-----------------------------------------------------------------------------------
def fillBodyDataLCPlanePlates(bodyIds, result, snEqual, geometry, analysis, userUnitLength, meshUnit):
	bodyData = {}
	if snEqual == 'Yes':
		userTemp = result.Properties["snClass"].Properties["snEqual"].Properties["temperature"].Value
		environment = result.Properties["snClass"].Properties["snEqual"].Properties["environment"].ValueString
		plateThickness = result.Properties["snClass"].Properties["snEqual"].Properties["plateThickness"].Value
		plateThickness = units.ConvertUnit(plateThickness, userUnitLength, meshUnit, "Length")
		strengthCoeff, ductilityCoeff = getStrengthAndDuctility(environment)
		for bodyId in bodyIds:
			thisMaterial = geometry.GeoEntityById(bodyId).Material
			emodule = findEmodule(userTemp, analysis, bodyId)
			bodyData[bodyId] = [environment, emodule, strengthCoeff, ductilityCoeff, plateThickness]
	else:
		table = result.Properties["snClass"].Properties["snEqual"].Properties["fatigueTable"]
		for row in xrange(table.RowCount):
			table.ActiveRow = row
			bodyId = bodyIds[row]
			environment = table.Properties["environmentColumn"].ValueString
			emodule = table.Properties["emoduleColumn"].Value
			plateThickness = table.Properties["plateThicknessColumn"].Value
			plateThickness = units.ConvertUnit(plateThickness, userUnitLength, meshUnit, "Length")
			strengthCoeff, ductilityCoeff = getStrengthAndDuctility(environment)
			bodyData[bodyId] = [environment, emodule, strengthCoeff, ductilityCoeff, plateThickness]
	return bodyData	