#\file lcScreening.py
#
#----------------------------------------------------------------------------------
#
#   Implemented controller for low cycle fatigue calculation in base material
#   
#----------------------------------------------------------------------------------
#----------------------------------------------------------------------------------
#
# author: Tone Wermundsen
# Last update: 18.12.2014
# ansys v15 ACT 
#
#-----------------------------------------------------------------------------------

#----------------------------------------------------------------------------------- 
#   Nested controller class that handles all evaluation behavior of the result object
#-----------------------------------------------------------------------------------
class LCScreeningController():
	def __init__(self, ExtAPI, result):
		self.ExtAPI = ExtAPI
		self.result = result

		displayOption = result.Properties["integrationPointResult"].Properties["displayOption"]
		setDisplayOption(result, displayOption)

	def evaluate(self, result, stepInfo, collector):
		result.Attributes['timeStamp'] = 0
		result.Attributes['Analysis_list'] = []
		ExtAPI.Log.WriteMessage("ready for LC Screening")
		
		#decs
		myAnalysis = result.Analysis 
		myMesh = myAnalysis.MeshData
		myGeometry = myAnalysis.GeoData
		myReader = myAnalysis.GetResultsData()

		#units
		meshUnit = myMesh.Unit 
		geoUnit = myGeometry.Assemblies[0].Unit
		userUnitTemp = myAnalysis.CurrentConsistentUnitFromQuantityName("Temperature")
		userUnitStress = myAnalysis.CurrentConsistentUnitFromQuantityName("Stress")
		userUnitLength = myAnalysis.CurrentConsistentUnitFromQuantityName("Length")

		#user input
		geoIds = result.Properties["scope"].Properties["geometry"].Value.Ids
		shellOption = result.Properties["scope"].Properties["shellOption"].ValueString
		resultView = result.Properties["resultView"].ValueString
		displayOption = result.Properties["integrationPointResult"].Properties["displayOption"].ValueString
		lcCombo = result.Properties["loadCaseCombo"].ValueString
		snEqual = result.Properties["snClass"].Properties["snEqual"].ValueString
		N0 = result.Properties["Advanced"].Properties["initialGuess"].Value
		lcTable = result.Properties["loadCaseCombo"].Properties["loadCaseTable"]


		bodyIds = getBodyIds(result)
		if len(bodyIds) == 0:
			MessageBox.Show(ExtAPI.UserInterface.MainWindow, "Unvalid input.", 'Error', MessageBoxType.Error, MessageBoxButtons.OK) 
			return
		bodyData = fillBodyDataLCScreening(bodyIds, result, snEqual, myGeometry, myAnalysis)		
		loadCase = fillLoadCaseInfo(lcCombo, myAnalysis, myReader, result, resultView, lcTable)

		summedLogN = {}; summedDamage = {}; summedN = {}
		for elementId in collector.Ids:
			summedLogN[elementId] = []
			summedDamage[elementId] = []
			summedN[elementId] = []

		excepter = 0	
		for lc in loadCase:
			loadCaseAnalysis = loadCase[lc][1]
			startTime = loadCase[lc][2]
			endTime = loadCase[lc][3]
			loadCaseReader = loadCaseAnalysis.GetResultsData()
			loadCaseResultItemStress = loadCaseReader.GetResult('EPTO')
			numberOfCycles = loadCase[lc][4]
			Shell = False

			resultComponentsT1 = {}; resultComponentsT2 = {}
			loadCaseReader.CurrentResultSet = loadCaseReader.ListTimeFreq.IndexOf(endTime)+1
			for elementId in collector.Ids:
				elementType = myMesh.ElementById(elementId).Type 
				if str(elementType) in ['kQuad8','kQuad4','kTri3','kTri6']:
					Shell = True
				else:
					Shell = False	
				resultComponentsT2[elementId] = getAllResultComponents_FatigueDNV(loadCaseResultItemStress, elementId, 1, shellOption, Shell)
			if startTime != 0:
				loadCaseReader.CurrentResultSet = loadCaseReader.ListTimeFreq.IndexOf(startTime)+1
				for elementId in collector.Ids:
					elementType = myMesh.ElementById(elementId).Type			
					if str(elementType) in ['kQuad8','kQuad4','kTri3','kTri6']:
						Shell = True 
					else:	
						Shell = False
					resultComponentsT1[elementId] = getAllResultComponents_FatigueDNV(loadCaseResultItemStress, elementId, 1, shellOption, Shell)
			else:
				for elementId in collector.Ids:
					resultComponentsT1[elementId] = [0.0]*len(resultComponentsT2[elementId])

			resultSet = {}		
			for elementId in collector.Ids:
				resultSet[elementId] = calculatePrincipal_FatigueDNV(resultComponentsT1[elementId], resultComponentsT2[elementId], False, 'nothing')	

			if resultView != 'Strain Range':
				logN = {}; damage = {}; N = {}
				N, excepter, highLife, errorLog = calculateLifeLowCycleNormal(myMesh, collector, resultSet, bodyData, N0)
				for elementId in collector.Ids:
					if resultView == 'Damage':
						damage[elementId] = calculateDamage(N[elementId], numberOfCycles)

				for elementId in collector.Ids:
					if lcCombo == 'No':
						summedN[elementId] = N[elementId]
						summedLogN[elementId] = map(math.log10, summedN[elementId])
						if resultView == 'Damage':
							summedDamage[elementId] = damage[elementId]
					else:
						if lc == 0:
							summedN[elementId] = map(operator.div, len(N[elementId])*(1, ), N[elementId])
							if resultView == 'Damage':
								summedDamage[elementId] = damage[elementId]
						if lc != 0:
							for i in xrange(len(N[elementId])):
								try:
									summedN[elementId][i] += 1/N[elementId][i]
								except:
									pass
								if resultView == 'Damage':
									summedDamage[elementId][i] += damage[elementId][i]	
						if lc == lcTable.RowCount-1:
							summedN[elementId] = map(operator.div, len(summedN[elementId])*(1, ), summedN[elementId])
							summedLogN[elementId] = map(math.log10, summedN[elementId])		

							
				if excepter != 0:
					MessageBox.Show(ExtAPI.UserInterface.MainWindow, "Initial Guess for iteration method Newton-Raphson is not within the methods convergence interval. Please try a different value as some results are not accurate.", 'Warning', MessageBoxType.Warning, MessageBoxButtons.OK)	
				if highLife != 0:
					ExtAPI.Application.LogWarning("Number of cycles to failure exceeds 1.0e+10 for some areas. Use results with caution")
				writeToFile(errorLog, myAnalysis)

		if resultView == 'Strain Range':
			for elementId in collector.Ids:
				collector.SetValues(elementId, resultSet[elementId])
		if resultView == 'Damage':
			for elementId in collector.Ids:
				collector.SetValues(elementId, summedDamage[elementId])
		if resultView == 'Cycles To Failure':
			for elementId in collector.Ids:
				collector.SetValues(elementId, summedN[elementId])	
		if resultView == 'Logarithm Of Cycles To Failure To Base 10':
			for elementId in collector.Ids:
				collector.SetValues(elementId, summedLogN[elementId])						

		result.Attributes['timeStamp'] = time.time()

#----------------------------------------------------------------------------------- 
#   Fills bodyData with information about each bodies environment and material properties
#-----------------------------------------------------------------------------------
def fillBodyDataLCScreening(bodyIds, result, snEqual, geometry, analysis):
	bodyData = {}
	if snEqual == 'Yes':
		userTemp = result.Properties["snClass"].Properties["snEqual"].Properties["temperature"].Value
		environment = result.Properties["snClass"].Properties["snEqual"].Properties["environment"].ValueString
		strengthCoeff, ductilityCoeff = getStrengthAndDuctility(environment)
		for bodyId in bodyIds:
			thisMaterial = geometry.GeoEntityById(bodyId).Material
			emodule = findEmodule(userTemp, analysis, bodyId)
			bodyData[bodyId] = [environment, emodule, strengthCoeff, ductilityCoeff]
	else:
		table = result.Properties["snClass"].Properties["snEqual"].Properties["fatigueTable"]
		for row in xrange(table.RowCount):
			table.ActiveRow = row
			bodyId = bodyIds[row]
			environment = table.Properties["environmentColumn"].ValueString
			emodule = table.Properties["emoduleColumn"].Value
			strengthCoeff, ductilityCoeff = getStrengthAndDuctility(environment)
			bodyData[bodyId] = [environment, emodule, strengthCoeff, ductilityCoeff]
	return bodyData			

