#\file mathHelpers.py
#
#----------------------------------------------------------------------------------
#
#   Implemented mathematical functions for use in fatigue calculation
#   
#----------------------------------------------------------------------------------
#----------------------------------------------------------------------------------
#
# author: Tone Wermundsen
# Last update: 01.12.2014
# ansys v15 ACT 
#
#-----------------------------------------------------------------------------------



#----------------------------------------------------------------------------------- 
#   input all extrapolated values to a hot spot, combines from the two distances
#   and returns a fulle extrapolated value stress/stran range_hotspot  
#-----------------------------------------------------------------------------------

def getCombinedExtValues(smallValues, largeValues, thickness):
    a = 0.5*thickness
    b = 1.5*thickness
    dhs = [[0.0,0.0,0.0],[0.0,0.0,0.0],[0.0,0.0,0.0]]
    for i in xrange(len(smallValues)):
        if -2 in smallValues[i]:
            pass
        else:
            for j in xrange(len(smallValues[i])):
                da = smallValues[i][j]
                db = largeValues[i][j]
                dhs[i][j] = da - ((a/(b-a))*(db-da)) 
    return dhs


#-----------------------------------------------------------------------------------  
#   finds point based on distance from hot spot in set direction
#-----------------------------------------------------------------------------------
def getExtCoordinates(mesh, hotspotId, directions, thickness):
    hotspot = mesh.NodeById(hotspotId)
    extCoordinates = []

    for direction in directions:
        x = hotspot.X + thickness*direction[0]
        y = hotspot.Y + thickness*direction[1]
        z = hotspot.Z + thickness*direction[2]
        extPoint = [x,y,z]
        extCoordinates.append(extPoint)
    return extCoordinates     

#-----------------------------------------------------------------------------------
#   checks if a given point is on within a definded line's boundary
#-----------------------------------------------------------------------------------
def pointIsOnLine(mesh, line, point):
    #the point is on the infinte line if: 
    #(pointX-startPointX)/(endPointX-startPOintX) = (pointY-startPointY)/(endPointY-startPointY) = (pointZ-startPointZ)/(endPointZ-startPointZ)
    #on line if lengthSE (lengthLine) = lengthSP + lengthPE 
    startPoint = mesh.NodeById(line[3])
    endPoint = mesh.NodeById(line[4])


    lineL = math.sqrt(line[0]**2 + line[1]**2 + line[2]**2)
    spL   = math.sqrt((point[0] - startPoint.X)**2  + (point[1] - startPoint.Y)**2 + (point[2] - startPoint.Z)**2)
    peL   = math.sqrt((endPoint.X - point[0])**2 + (endPoint.Y - point[1])**2 + (endPoint.Z - point[2])**2)
    if almostEqual(lineL, (spL+peL)):#(lineL == (spL + peL)):
        return True 
    return False  

#-----------------------------------------------------------------------------------
#   Rrtates a vector angle degress aboun an axis w
#-----------------------------------------------------------------------------------
def rodriguesRotateVector(w, u, angle):
    v = u 
    cosAngleXV = map(operator.mul, len(v)*(math.cos(angle), ), v)
    wXv = crossProduct(w,v)
    sinAngleXwXv = map(operator.mul, len(v)*(math.sin(angle), ), wXv)
    wDotv = dotProduct(w,v)
    wXwDotV = map(operator.mul, len(w)*(wDotv, ), w)
    OneMinCosAngle = 1- math.cos(angle)
    last = map(operator.mul, len(wXwDotV)*(OneMinCosAngle, ), wXwDotV)
    v = map(operator.add, cosAngleXV,sinAngleXwXv) 
    v = map(operator.add, v, last)
    return v

#-----------------------------------------------------------------------------------
#   returns normal vector based on 3 points that define a plane
#-----------------------------------------------------------------------------------
def getNormalVector(center, node0, node1):
    line1 = [center[0]-node0.X, center[1]-node0.Y, center[2]-node0.Z]
    line2 = [center[0]-node1.X, center[1]-node1.Y, center[2]-node1.Z]

    normal = crossProduct(line1, line2)
    return normal

#-----------------------------------------------------------------------------------
#   find interpolated value in a 4-noded element with values on all 4 cornernodes
#-----------------------------------------------------------------------------------
def findInterpolatedValue(mesh, elements, locations, nodeValues, t2, lc):
    myValues = [[0.0]*3]*len(elements)
    for i in xrange(len(elements)):
        elementId = elements[i]
        pointToFind = locations[i]
        if elementId != -2:
            element = mesh.ElementById(elementId)
            elementCenter = element.Centroid
            cornerNodes = element.CornerNodeIds
            if element.CornerNodeCount == 4:
                node0 = mesh.NodeById(cornerNodes[0])
                node1 = mesh.NodeById(cornerNodes[1])
                node2 = mesh.NodeById(cornerNodes[2])
                node3 = mesh.NodeById(cornerNodes[3])
                elementXaxis = [node1.X-node0.X, node1.Y-node0.Y, node1.Z-node0.Z]
                elementZaxis = getNormalVector(elementCenter, node0, node1)
                elementYaxis = crossProduct(elementXaxis, elementZaxis)
                #convert system
                node0Conv = convertPoint([node0.X, node0.Y, node0.Z], elementCenter, elementXaxis, elementYaxis, elementZaxis)
                node1Conv = convertPoint([node1.X, node1.Y, node1.Z], elementCenter, elementXaxis, elementYaxis, elementZaxis)
                node2Conv = convertPoint([node2.X, node2.Y, node2.Z], elementCenter, elementXaxis, elementYaxis, elementZaxis)
                node3Conv = convertPoint([node3.X, node3.Y, node3.Z], elementCenter, elementXaxis, elementYaxis, elementZaxis)
                pointToFindConv = convertPoint(pointToFind, elementCenter, elementXaxis, elementYaxis, elementZaxis)

                e,n = findMappedValues(node0Conv, node1Conv, node2Conv, node3Conv, pointToFindConv)
                node0Result = nodeValues[lc, cornerNodes[0], elementId, t2]
                node1Result = nodeValues[lc, cornerNodes[1], elementId, t2]
                node2Result = nodeValues[lc, cornerNodes[2], elementId, t2]
                node3Result = nodeValues[lc, cornerNodes[3], elementId, t2]


                N1 = (1./4.)*(1.-e)*(1.-n)  
                N2 = (1./4.)*(1.+e)*(1.-n)
                N3 = (1./4.)*(1.+e)*(1.+n)
                N4 = (1./4.)*(1.-e)*(1.+n)

                bottom = N1*node0Result[0] + N2*node1Result[0] + N3*node2Result[0] + N4*node3Result[0]
                top = N1*node0Result[1] + N2*node1Result[1] + N3*node2Result[1] + N4*node3Result[1]
                middle = N1*node0Result[2] + N2*node1Result[2] + N3*node2Result[2] + N4*node3Result[2]

                myValues[i] = [bottom, top, middle]
            else:
                pass
        else:
            myValues[i] = [-2,-2,-2]        
    return myValues

#-----------------------------------------------------------------------------------
#   mapping of a isoparamedtric polygon to a perfect square
#-----------------------------------------------------------------------------------
def findMappedValues(node0, node1, node2, node3, point):
    #x:
    A = node0[0] + node1[0] + node2[0] + node3[0]
    B = -node0[0] - node1[0] + node2[0] + node3[0]
    C = -node0[0] + node1[0] + node2[0] - node3[0]
    D = node0[0] - node1[0] + node2[0] - node3[0]
    S = node0[1] + node1[1] + node2[1] + node3[1]
    Q = -node0[1] - node1[1] + node2[1] + node3[1]
    R = -node0[1] + node1[1] + node2[1] - node3[1]
    T = node0[1] - node1[1] + node2[1] - node3[1]
    a = (C*T) + (D*R)
    b = (A*T) - (B*R) + (C*Q) - (D*4*point[1]) - (D*S) - (4.*T*point[0])
    c = (A*Q) + (B*4.*point[1]) - (B*S) + (Q*4.*point[0])

    if a != 0:
        e1 = (-b + math.sqrt((b**2) -(4*a*c)))/(2*a)
        e2 = (-b - math.sqrt((b**2) - (4*a*c)))/(2*a)
        try:
            n1 = ( (4.*point[1]) - S - (R*e1))/( Q + (T*e1))
        except:
            n1 = 0
        try:
            n2 = ( (4.*point[1]) - S - (R*e2))/( Q + (T*e2))
        except:
            n2 = 0  
        if  (-1 < e1) and (e1 < 1):
            if  (-1 < n1) and (n1 < 1): 
                e = e1
                n = n1
            else:
                e = e2
                n = n2  
        else:
            e = e2
            n = n2          
    else:
        e = (-(A*Q) - (B*4.*point[1]) + (B*S) + (Q*4.*point[0]))/( (A*T) - (B*R) + (Q*C) + (D*4.*point[1]) - (D*S) - (T*4.*point[0]))
        try:
            n = ( (4.*point[1]) - S - (R*e))/( Q + (T*e))
        except:
            n = 0
    return e,n 

#-----------------------------------------------------------------------------------
#   converts a points coordinate based on a new defined coordinatesystem
#-----------------------------------------------------------------------------------
def convertPoint(pointToConvert, csOrigin, csXdir, csYDir, csZDir):
    transformationMatrix = [csXdir, csYDir, csZDir]
    displacedPoint = [pointToConvert[0]-csOrigin[0], pointToConvert[1]-csOrigin[1], pointToConvert[2]-csOrigin[2]]
    convertedPoint = multiplyMatrixAndVector(transformationMatrix, displacedPoint)
    return [convertedPoint[0], convertedPoint[1]]

#-----------------------------------------------------------------------------------
#   multiplies a maxtrix and vector
#-----------------------------------------------------------------------------------
def multiplyMatrixAndVector(matrix, vector):
    resultvector = [0 for x in xrange(len(vector))]
    for i in xrange(0,len(matrix)):
        for j in xrange(0,len(matrix[0])): 
                resultvector[i] += matrix[i][j] * vector[j] 
    return resultvector

#-----------------------------------------------------------------------------------
#   takes a vector and returns it's unitvector
#-----------------------------------------------------------------------------------
def unitVector(vector):     
    summ = 0
    unitVector = []
    for item in vector:
        summ += item**2
    vectorLength = math.sqrt(summ)
    if vectorLength == 0:
        unitVector = vector
    if vectorLength != 0:
        for i in xrange(len(vector)):
            unitVector.append(vector[i]/vectorLength)    
    return unitVector  

#-----------------------------------------------------------------------------------
#   returns the cross product of two 3dimensjonal vectors (x,y,z)
#-----------------------------------------------------------------------------------
def crossProduct(v1,v2):
    c = [v1[1]*v2[2] - v1[2]*v2[1], 
        v1[2]*v2[0] - v1[0]*v2[2], 
        v1[0]*v2[1] - v1[1]*v2[0]]
    return c

#-----------------------------------------------------------------------------------
#   ray-casting algorithm to check if a point is inside a polygon
#-----------------------------------------------------------------------------------
def rayCasting(mesh, line, startNode, endNode, rayStart, rayEnd):
    start = mesh.NodeById(startNode)
    end = mesh.NodeById(endNode)

    lineStart = [start.X, start.Y, start.Z]
    lineEnd = [end.X, end.Y, end.Z]

    l1 = [lineEnd[0] - lineStart[0], lineEnd[1] - lineStart[1], lineEnd[2] - lineStart[2]]
    l2 = [rayEnd[0] - rayStart[0], rayEnd[1] - rayStart[1], rayEnd[2] - rayStart[2]]
    if l2 == [0,0,0]:
        l2 = [rayEnd[0] - rayStart[0]*0.8, rayEnd[1] - rayStart[1]*0.8, rayEnd[2] - rayStart[2]*0.08]
    l3 = [lineStart[0] - rayStart[0], lineStart[1] - rayStart[1], lineStart[2] - rayStart[2]]

    u1 = unitVector(l1)
    u2 = unitVector(l2)

    m1 = crossProduct(lineStart, lineEnd)
    m2 = crossProduct(rayStart, rayEnd)
    #starter eller ender likt
    if lineStart == rayStart or lineStart == rayEnd or lineEnd == rayStart or lineEnd == rayEnd:
        return 1
    #parallelle, enten er de parallelle likt (deler en segment) eller saa deler de ikke 
    elif vectorLength(crossProduct(u1,u2)) == 0:
        if pointIsOnLine(mesh, line, rayEnd):
            return 1
        else:
            return 0
    #skew: de er ikke i samme plan
    elif fourPointVolume(lineStart, lineEnd, rayStart, rayEnd) != 0:
        return 0
    #de er i samme plan, er ikke parallelle: ergo de maa krysse et sted. Enten krysser de paa elemLine segment, eller ikke      
    else:
        u = crossProduct(l3, u2)
        v = crossProduct(u1,u2)
        a = vectorLength(u)/vectorLength(v)
        myPointX = lineStart[0] + a*u1[0]
        myPointY = lineStart[1] + a*u1[1]
        myPointZ = lineStart[2] + a*u1[2]
        myPoint = [myPointX, myPointY, myPointZ]
        if pointIsOnLine(mesh, line, myPoint) == True:
            if pointOnRay(l2, rayStart, rayEnd, myPoint) == True:
                return 1
            else:
                return 0    
        else:   
            return 0

#-----------------------------------------------------------------------------------
#   volume defined by 4 points
#-----------------------------------------------------------------------------------
def fourPointVolume(p1,p2,p3,p4):
    l1 = [p1[0]-p2[0], p1[1]-p2[1], p1[2]-p2[2]]
    l2 = [p2[0]-p3[0], p2[1]-p3[1], p2[2]-p3[2]]    
    l3 = [p3[0]-p4[0], p3[1]-p4[1], p3[2]-p4[2]]
    matrix = [l1,l2,l3]
    det = getMatrixDeterminant(matrix)
    volume = abs(det)/6

    return volume
#-----------------------------------------------------------------------------------
#   returns true if an integer is odd
#-----------------------------------------------------------------------------------
def isOdd(number):
    if number % 2 == 0:
        return False
    if number == 0:
        return False    
    else:
        return True     

#-----------------------------------------------------------------------------------
#   returns a vectors length
#-----------------------------------------------------------------------------------
def vectorLength(vector):
    return math.sqrt((vector[0])**2 + (vector[1])**2 + (vector[2])**2)  

#-----------------------------------------------------------------------------------
#   iteration method newton-raphson with relaxation
#-----------------------------------------------------------------------------------
def newtonRaphson(elementId, cpt, deltaStrain, emodule, strengthCoeff, ductilityCoeff, N0):
    # n1 = n0 - f(n0)/dF(n0)
    delta = 1e-4
    itMax = 100
    itera = 0
    diff = 1.
    excepter = 0
    highLife = 0
    dictFileLog = []
    #if straion range is so small that results, N, becomes too high to be valid
    if func(deltaStrain, emodule, strengthCoeff, ductilityCoeff, 10e9) >= 0.:
        N1 = 10e9
        highLife += 1
    else:
        if func(deltaStrain, emodule, strengthCoeff, ductilityCoeff, N0) > 0:
            N0 = N0*10  
        elif func(deltaStrain, emodule, strengthCoeff, ductilityCoeff, N0) < 0:
            #guess is to high
            if func(deltaStrain, emodule, strengthCoeff, ductilityCoeff, N0/10) < 0:
                #guess is still too high
                N0 = N0/100
            else:   
                N0 = N0/10

        while abs(diff) > delta:
            itera += 1
            #when max iter is reached
            if itera > itMax:
                excepter += 1
                try:
                    dictFileLog = [elementId, cpt, deltaStrain, diff, N1, itera, func(deltaStrain, emodule, strengthCoeff, ductilityCoeff, N0), dFunc(emodule, strengthCoeff, ductilityCoeff, N0)]
                except:
                    dictFileLog = [elementId, cpt, deltaStrain, diff, N1, itera, N0, 0]
                break
            N1 = N0 - (func(deltaStrain, emodule, strengthCoeff, ductilityCoeff, N0)/dFunc(emodule, strengthCoeff, ductilityCoeff, N0)) 
            testeriter = 0
            while N1 < 0:
                relaxer = 0.1-((testeriter*0.1)/100)
                testeriter += 1
                #relaxation
                N1 = N0 - relaxer*(func(deltaStrain, emodule, strengthCoeff, ductilityCoeff, N0)/dFunc(emodule, strengthCoeff, ductilityCoeff, N0)) 
            diff = N1-N0
            N0 = N1     

    return abs(N1), excepter, highLife, dictFileLog 

#-----------------------------------------------------------------------------------
#   function for calculating N, low cycle: 
#               strainRange/2 = (strengthcoeff)*2N^-0.1/E + ductilitycoeff*2N^-0.43
#-----------------------------------------------------------------------------------
def func(deltaStrain, emodule, strengthCoeff, ductilityCoeff, N):
    return ((strengthCoeff*((2.*N)**(-0.1))/emodule) + (ductilityCoeff*((2.*N)**(-0.43))) - (deltaStrain/2.))

#-----------------------------------------------------------------------------------
#   derivative of function for calclating N, low cycle
#-----------------------------------------------------------------------------------
def dFunc(emodule, strengthCoeff, ductilityCoeff, N):
    return ((-0.1*strengthCoeff*(2.**(-0.1))*((N)**(-1.1))/emodule) - (0.43*ductilityCoeff*(2.**(-0.43))*((N)**(-1.43))))

#-----------------------------------------------------------------------------------
#   calculates damage
#-----------------------------------------------------------------------------------
def calculateDamage(N, numberOfCycles):
    damage = map(operator.div, len(N)*(numberOfCycles, ), N)
    return damage

#-----------------------------------------------------------------------------------
#   returns the dot product of two 3dimensjonal vectors (x,y,z)
#-----------------------------------------------------------------------------------
def dotProduct(u,v):
    c = u[0]*v[0] + u[1]*v[1] + u[2]*v[2]
    return c


#--------------------------------
#
# EigenValue_FatigueDNV
#		input: low leveled stress/strain matrix: xx,yy,zz,0.5xy,0.5xz,0.5yz
#		output: principal stresses s1,s2 and s3
#		NOTE: with small shear stress/strain s1,s2,s3 = sorted(xx,yy,zz)
#		otherwise, matrix operations is done
#
#--------------------------------
"""Mathematical functions"""
#From multiRstList by Magnus Gustafsson EDRMedeso 2014-06-13 Update for R15 
def eigenValue_FatigueDNV(a, b, c, d, e, f):
    principal = {}
    # Check if matrix is diagonal
    abc = abs(a)+abs(b)+abs(c)
    if abc == 0.0 : abc = 1.e-30
    if (abs(d)+abs(e)+abs(f))/abc > 1.e-4 :
        # Polynomial reduction
        A = -(a+b+c)
        B = a*b+a*c+b*c-d*d-e*e-f*f
        C = d*d*c+f*f*a+e*e*b-2.*d*e*f-a*b*c
        p = B-A*A/3.
        q = C-A*B/3.+2.*A*A*A/27.
        R = math.sqrt(math.fabs(p)/3.)
        if q < 0: R = -R
        z = q/(2.*R*R*R)
        if z < -1. : z = -1.
        elif z > 1.: z = 1.
        phi = math.acos(z)
        S1 = -2.*R*math.cos(phi/3.)-A/3.
        S2 = -2.*R*math.cos(phi/3.+2.*math.pi/3.)-A/3.
        S3 = -2.*R*math.cos(phi/3.+4.*math.pi/3.)-A/3.
    else:
        S1 = a
        S2 = b
        S3 = c
    # Sort the eigen values so that S1>S2>S3
    absMaxPrin=max(abs(S1), abs(S2), abs(S3))
    principal=[absMaxPrin,S1, S2, S3]
    principal.sort(reverse=True)
    return (principal) 	

#-----------------------------------------------------------------------------------
#   returns determinant of a matrix
#-----------------------------------------------------------------------------------
def getMatrixDeterminant(matrix):
    matLength = len(matrix)
    if(matLength > 2):
        i = 1
        t = 0
        sum = 0
        while t<=matLength-1:
            d = {}
            t1 = 1
            while t1<=matLength-1:
                m = 0
                d[t1] = []
                while m<=matLength-1:
                    if m==t:
                        u = 0
                    else:
                        d[t1].append(matrix[t1][m])
                    m += 1
                t1 += 1
            matrix1 = [d[x] for x in d]
            sum = sum +i*(matrix[0][t]*getMatrixDeterminant(matrix1))
            i = i*(-1)
            t += 1  
        return sum
    else:
        determinant = (matrix[0][0]*matrix[1][1]-matrix[0][1]*matrix[1][0])                          
        return determinant

#-----------------------------------------------------------------------------------
#   checks if  a point is on a rayLine
#-----------------------------------------------------------------------------------
def pointOnRay(ray, rayStart, rayEnd, myPoint):
    lineL = math.sqrt(ray[0]**2 + ray[1]**2 + ray[2]**2)
    spL = math.sqrt((myPoint[0]-rayStart[0])**2 + (myPoint[1]-rayStart[1])**2 + (myPoint[2]-rayStart[2])**2)
    peL = math.sqrt((rayEnd[0]-myPoint[0])**2 + (rayEnd[1] - myPoint[1])**2 + (rayEnd[2] - myPoint[2])**2)

    if almostEqual(lineL, (spL + peL)):
        return True
    else:
        return False 

#-----------------------------------------------------------------------------------
#   checks if two values are alomost equal
#-----------------------------------------------------------------------------------
def almostEqual(x,y):
    if x  ==  y:
        return True
    elif ((y-(y/10000000)) < x) and (x < (y+(y/10000000))):
        return True 
    elif ((y-(y/10000000)) > x) and (x > (y+(y/10000000))): 
        return True
    else:
        return False 